from opennem.pipelines.nem.opennem import NemwebUnitScadaOpenNEMStorePipeline
from opennem.spiders.nemweb import NemwebSpider


class NemwebCurrentDispatchIS(NemwebSpider):
    name = "au.nem.current.dispatch_is"
    start_url = "http://nemweb.com.au/Reports/Current/DispatchIS_Reports/"
    limit = 0

    pipelines_extra = set([NemwebUnitScadaOpenNEMStorePipeline])


class NemwebArchiveDispatchIS(NemwebSpider):
    name = "au.nem.archive.dispatch_is"
    start_url = "http://nemweb.com.au/Reports/Archive/DispatchIS_Reports/"
    limit = 0

    pipelines_extra = set([NemwebUnitScadaOpenNEMStorePipeline])

    # Archives tend to contain large zips of embedded zips so throttle
    # to limit memory use
    custom_settings = {
        "CONCURRENT_REQUESTS": 1,
        "CONCURRENT_ITEMS": 1,
    }



"""
    Parse facility codes facility ids

"""

WEM_CODE_SUFFIXES = ["plant"]


def parse_wem_facility_code(facility_code: str) -> str:
    """
        Parses WEM facility codes into station_code and facility_code

        If the suffix is in the map include it

        ex. PINJAR_GT7 becomes PINJAR and PINJAR_GT7
    """

    if not type(facility_code) is str:
        raise Exception("Invalid facilty code: {}".format(facility_code))

    if not facility_code:
        # Exception?
        return None

    if "_" not in facility_code:
        return facility_code

    comp = facility_comp = facility_code.split("_")

    num_comp = len(comp)

    if comp[num_comp - 1].lower() not in WEM_CODE_SUFFIXES:
        facility_comp = comp[: num_comp - 1]

    return "_".join(facility_comp)



import logging

from sqlalchemy.dialects.postgresql import insert

from opennem.core.networks import network_from_state
from opennem.db import SessionLocal, get_database_engine
from opennem.db.models.opennem import Facility, FacilityScada
from opennem.importer.apvi import ROOFTOP_CODE
from opennem.utils.dates import parse_date
from opennem.utils.pipelines import check_spider_pipeline

logger = logging.getLogger(__name__)


STATE_POSTCODE_PREFIXES = {
    "NSW": "2",
    "VIC": "3",
    "QLD": "4",
    "SA": "5",
    "WA": "6",
    "TAS": "7",
    "NT": "0",
}


class APVIStoreData(object):
    @check_spider_pipeline
    def process_item(self, item, spider=None):
        if "records" not in item:
            logger.error("Invalid return response")

        records = item["records"]

        is_latest = False
        record_date = None

        if "meta" in item:
            if "is_latest" in item["meta"]:
                is_latest = item["meta"]["is_latest"]

            if "record_date" in item["meta"]:
                record_date = item["meta"]["record_date"]

        if "postcode" not in records:
            logger.error("No postcode data")

        if "installations" not in records:
            logger.error("No postcode data")

        if "postcodeCapacity" not in records:
            logger.error("No postcode capacity data")

        postcode_gen = records["postcode"]
        postcode_capacity = records["postcodeCapacity"]
        installations = records["installations"]

        engine = get_database_engine()
        session = SessionLocal()

        records_to_store = []

        for record in postcode_gen:
            for state, prefix in STATE_POSTCODE_PREFIXES.items():
                facility_code = "{}_{}".format(ROOFTOP_CODE, state.upper())

                interval_time = parse_date(
                    record["ts"], dayfirst=False, yearfirst=True
                )

                network = network_from_state(state)
                interval_time = interval_time.astimezone(
                    network.get_timezone()
                )

                generated = sum(
                    [
                        v
                        for k, v in record.items()
                        if k.startswith(prefix) and v
                    ]
                )

                if not generated:
                    continue

                __record = {
                    "network_id": network.code,
                    "trading_interval": interval_time,
                    "facility_code": facility_code,
                    "generated": generated,
                }

                records_to_store.append(__record)

        STATE_CAPACITIES = {}

        if is_latest:
            # temporariy only run getting capacities on latest
            logger.info("Updating capacities on %s", record_date)

            for postcode_prefix, capacity_val in postcode_capacity.items():
                for state, prefix in STATE_POSTCODE_PREFIXES.items():
                    if state not in STATE_CAPACITIES:
                        STATE_CAPACITIES[state] = 0

                    if postcode_prefix.startswith(prefix):
                        STATE_CAPACITIES[state] += capacity_val

            for state, state_capacity in STATE_CAPACITIES.items():
                facility_code = "{}_{}".format(ROOFTOP_CODE, state.upper())

                state_facility: Facility = session.query(Facility).filter_by(
                    code=facility_code
                ).one_or_none()

                if not state_facility:
                    raise Exception(
                        "Could not find rooftop facility for %s", facility_code
                    )

                state_facility.capacity_registered = state_capacity

                if state.lower() in installations:
                    state_number_units = installations[state.lower()]
                    state_facility.unit_number = state_number_units

                session.add(state_facility)
                session.commit()

        if len(records_to_store) < 1:
            return 0

        stmt = insert(FacilityScada).values(records_to_store)
        stmt.bind = engine
        stmt = stmt.on_conflict_do_update(
            index_elements=["trading_interval", "network_id", "facility_code"],
            set_={"generated": stmt.excluded.generated,},
        )

        try:
            session.execute(stmt)
            session.commit()
        except Exception as e:
            logger.error("Error: {}".format(e))
        finally:
            session.close()

        return len(records_to_store)



from pydantic import BaseModel


class BaseConfig(BaseModel):
    class Config:
        orm_mode = True
        anystr_strip_whitespace = True

        arbitrary_types_allowed = True
        validate_assignment = True

        json_encoders = {
            # datetime: lambda v: v.isotime(),
            # Decimal: lambda v: float(v),
        }



"""
    settings files - read settings from env and from config.yml



"""


import os
from typing import Optional

from .schema import OpennemSettings

MODULE_DIR = os.path.dirname(__file__)

ENV = os.getenv("ENV", default="development")

MYSQL_HOST_URL = os.getenv("MYSQL_HOST_URL", default=False)

# settings default loader

settings: Optional[OpennemSettings] = None

if ENV != "development" and os.path.isfile(".env.production"):
    settings = OpennemSettings(_env_file=".env.production")
else:
    settings = OpennemSettings()



"""
    OpenNEM primary schema adapted to support multiple energy sources

    Currently supported:

    - NEM
    - WEM
"""

from decimal import Decimal
from typing import Optional

from dictalchemy import DictableModel
from geoalchemy2 import Geometry
from shapely import wkb
from sqlalchemy import (
    JSON,
    Boolean,
    Column,
    Date,
    DateTime,
    Enum,
    ForeignKey,
    Index,
    Integer,
    LargeBinary,
    Numeric,
    Text,
    func,
    text,
)
from sqlalchemy.dialects.postgresql import TIMESTAMP
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship

from opennem.core.dispatch_type import DispatchType
from opennem.core.oid import get_ocode, get_oid

Base = declarative_base(cls=DictableModel)
metadata = Base.metadata


class BaseModel(object):
    """
        Base model for both NEM and WEM

    """

    created_by = Column(Text, nullable=True)
    # updated_by = Column(Text, nullable=True)
    # processed_at = Column(DateTime(timezone=True), nullable=True)
    created_at = Column(DateTime(timezone=True), server_default=func.now())
    updated_at = Column(DateTime(timezone=True), onupdate=func.now())


class FuelTech(Base, BaseModel):
    __tablename__ = "fueltech"

    code = Column(Text, primary_key=True)
    label = Column(Text, nullable=True)
    renewable = Column(Boolean, default=False)

    facilities = relationship("Facility")


class Network(Base, BaseModel):
    __tablename__ = "network"

    code = Column(Text, primary_key=True)
    country = Column(Text, nullable=False)
    label = Column(Text, nullable=True)
    timezone = Column(Text, nullable=False)
    timezone_database = Column(Text, nullable=True)
    offset = Column(Integer, nullable=True)
    interval_size = Column(Integer, nullable=False)


class FacilityStatus(Base, BaseModel):
    __tablename__ = "facility_status"

    code = Column(Text, primary_key=True)
    label = Column(Text)


class Participant(Base, BaseModel):
    __tablename__ = "participant"

    id = Column(Integer, autoincrement=True, nullable=False, primary_key=True,)

    code = Column(Text, unique=True, index=True)
    name = Column(Text)
    network_name = Column(Text)
    network_code = Column(Text)
    country = Column(Text)
    abn = Column(Text)

    approved = Column(Boolean, default=False)
    approved_by = Column(Text)
    approved_at = Column(DateTime(timezone=True), nullable=True)


class Photo(Base):
    __tablename__ = "photo"

    id = Column(Integer, autoincrement=True, nullable=False, primary_key=True)

    station_id = Column(
        Integer,
        ForeignKey("station.id", name="fk_photos_station_id"),
        nullable=True,
    )
    station = relationship("Station", back_populates="photos")

    name = Column(Text)
    mime_type = Column(Text)
    original_url = Column(Text, nullable=True)
    data = Column(LargeBinary, nullable=True)
    width = Column(Integer)
    height = Column(Integer)

    license_type = Column(Text, nullable=True)
    license_link = Column(Text, nullable=True)
    author = Column(Text, nullable=True)
    author_link = Column(Text, nullable=True)

    processed = Column(Boolean, default=False)
    processed_by = Column(Text)
    processed_at = Column(DateTime(timezone=True), nullable=True)

    approved = Column(Boolean, default=False)
    approved_by = Column(Text)
    approved_at = Column(DateTime(timezone=True), nullable=True)

    @hybrid_property
    def photo_url(self) -> Optional[str]:
        if self.name:
            return "https://photos.opennem.org.au/{}".format(self.name)

        return None


class BomStation(Base):
    __tablename__ = "bom_station"

    __table_args__ = (
        Index("idx_bom_station_geom", "geom", postgresql_using="gist"),
    )

    code = Column(Text, primary_key=True)
    state = Column(Text)
    name = Column(Text)
    name_alias = Column(Text, nullable=True)
    registered = Column(Date)

    # priority from 1-5
    priority = Column(Integer, default=5)
    is_capital = Column(Boolean, default=False)

    website_url = Column(Text, nullable=True)
    feed_url = Column(Text, nullable=True)

    observations = relationship("BomObservation")
    altitude = Column(Integer, nullable=True)

    geom = Column(Geometry("POINT", srid=4326, spatial_index=False))

    @hybrid_property
    def lat(self) -> Optional[float]:
        if self.geom:
            return wkb.loads(bytes(self.geom.data)).y

        return None

    @hybrid_property
    def lng(self) -> Optional[float]:
        if self.geom:
            return wkb.loads(bytes(self.geom.data)).x

        return None


class BomObservation(Base):
    __tablename__ = "bom_observation"

    observation_time = Column(DateTime, primary_key=True)

    station_id = Column(
        Text,
        ForeignKey("bom_station.code", name="fk_bom_observation_station_code"),
        primary_key=True,
    )
    station = relationship("BomStation")
    temp_apparent = Column(Numeric)
    temp_air = Column(Numeric)
    press_qnh = Column(Numeric)
    wind_dir = Column(Text, nullable=True)
    wind_spd = Column(Numeric)
    wind_gust = Column(Numeric)
    humidity = Column(Numeric, nullable=True)
    cloud = Column(Text, nullable=True)
    cloud_type = Column(Text, nullable=True)


class Location(Base):
    __tablename__ = "location"

    __table_args__ = (
        Index("idx_location_geom", "geom", postgresql_using="gist"),
        Index("idx_location_boundary", "boundary", postgresql_using="gist"),
    )

    id = Column(Integer, autoincrement=True, nullable=False, primary_key=True)

    # station_id = Column(Integer, ForeignKey("station.id"))

    # @TODO sort out this join based on this lateral query ..

    #  select l.id, l.locality, l.state, closest_station.state, closest_station.code, closest_station.dist from location l
    #  left join lateral (
    # 	select code, state, ST_Distance(l.geom, bom_station.geom) / 1000 as dist from bom_station order by l.geom <-> bom_station.geom limit 1
    #  ) AS closest_station on TRUE;

    # weather_station = relationship(
    #     "BomStation",
    #     primaryjoin="func.ST_ClosestPoint(remote(BomStation.geom), foreign(Location.geom))",
    #     viewonly=True,
    #     uselist=True,
    #     lazy="joined",
    # )

    address1 = Column(Text)
    address2 = Column(Text)
    locality = Column(Text)
    state = Column(Text)
    postcode = Column(Text, nullable=True)

    revisions = relationship("Revision", lazy="joined")

    # Geo fields
    place_id = Column(Text, nullable=True, index=True)
    geocode_approved = Column(Boolean, default=False)
    geocode_skip = Column(Boolean, default=False)
    geocode_processed_at = Column(DateTime, nullable=True)
    geocode_by = Column(Text, nullable=True)
    geom = Column(Geometry("POINT", srid=4326, spatial_index=False))
    boundary = Column(Geometry("MULTIPOLYGON", srid=4326, spatial_index=False))

    @hybrid_property
    def lat(self) -> Optional[float]:
        if self.geom:
            return wkb.loads(bytes(self.geom.data)).y

        return None

    @hybrid_property
    def lng(self) -> Optional[float]:
        if self.geom:
            return wkb.loads(bytes(self.geom.data)).x

        return None


class Station(Base, BaseModel):
    __tablename__ = "station"

    # __table_args__ = (
    #     UniqueConstraint(
    #         "customer_id", "location_code", name="_customer_location_uc"
    #     ),
    # )

    def __str__(self):
        return "{} <{}>".format(self.name, self.code)

    def __repr__(self):
        return "{} {} <{}>".format(self.__class__, self.name, self.code)

    id = Column(Integer, autoincrement=True, nullable=False, primary_key=True,)

    participant_id = Column(
        Integer,
        ForeignKey("participant.id", name="fk_station_participant_id"),
        nullable=True,
    )
    participant = relationship("Participant")

    location_id = Column(
        Integer,
        ForeignKey("location.id", name="fk_station_location_id"),
        nullable=True,
    )
    location = relationship("Location", lazy="joined", innerjoin=True)

    facilities = relationship("Facility", lazy="joined", innerjoin=True)

    revisions = relationship("Revision")

    photos = relationship("Photo")

    code = Column(Text, index=True, nullable=True)
    name = Column(Text)

    # wikipedia links
    description = Column(Text, nullable=True)
    wikipedia_link = Column(Text, nullable=True)
    wikidata_id = Column(Text, nullable=True)

    # Original network fields
    network_code = Column(Text, index=True)
    network_name = Column(Text)

    approved = Column(Boolean, default=False)
    approved_by = Column(Text)
    approved_at = Column(DateTime(timezone=True), nullable=True)

    @hybrid_property
    def network(self) -> Optional[Network]:
        """
            Return the network from the facility

        """
        if not self.facilities or not len(self.facilities) > 0:
            return None

        return self.facilities[0].network

    @hybrid_property
    def capacity_registered(self) -> Optional[int]:
        """
            This is the sum of registered capacities for all units for
            this station

        """
        cap_reg = None

        for fac in self.facilities:
            if (
                fac.capacity_registered
                and type(fac.capacity_registered) in [int, float, Decimal]
                and fac.status_id
                in ["operating", "committed", "commissioning"]
                and fac.dispatch_type == DispatchType.GENERATOR
                and fac.active
            ):
                if not cap_reg:
                    cap_reg = 0

                cap_reg += fac.capacity_registered

        if cap_reg:
            cap_reg = round(cap_reg, 2)

        return cap_reg

    @hybrid_property
    def capacity_aggregate(self) -> Optional[int]:
        """
            This is the sum of aggregate capacities for all units

        """
        cap_agg = None

        for fac in self.facilities:
            if (
                fac.capacity_aggregate
                and type(fac.capacity_aggregate) in [int, float, Decimal]
                and fac.status_id
                in ["operating", "committed", "commissioning"]
                and fac.dispatch_type == DispatchType.GENERATOR
                and fac.active
            ):
                if not cap_agg:
                    cap_agg = 0

                cap_agg += fac.capacity_aggregate

        if cap_agg:
            cap_agg = round(cap_agg, 2)

        return cap_agg

    @hybrid_property
    def oid(self) -> str:
        return get_oid(self)

    @hybrid_property
    def ocode(self) -> str:
        return get_ocode(self)


class Facility(Base, BaseModel):
    __tablename__ = "facility"

    def __str__(self):
        return "{} <{}>".format(self.code, self.fueltech_id)

    def __repr__(self):
        return "{} {} <{}>".format(self.__class__, self.code, self.fueltech_id)

    id = Column(Integer, autoincrement=True, nullable=False, primary_key=True,)

    network_id = Column(
        Text,
        ForeignKey("network.code", name="fk_station_network_code"),
        nullable=False,
    )
    network = relationship("Network")

    fueltech_id = Column(
        Text,
        ForeignKey("fueltech.code", name="fk_facility_fueltech_id"),
        nullable=True,
    )
    fueltech = relationship(
        "FuelTech", back_populates="facilities", lazy="joined", innerjoin=True
    )

    status_id = Column(
        Text,
        ForeignKey("facility_status.code", name="fk_facility_status_code"),
    )
    status = relationship("FacilityStatus", lazy="joined", innerjoin=True)

    station_id = Column(
        Integer,
        ForeignKey("station.id", name="fk_station_status_code"),
        nullable=True,
    )
    # station = relationship("Station", back_populates="facilities")

    revisions = relationship("Revision")

    # DUID but modified by opennem as an identifier
    code = Column(Text, index=True)

    # Network details
    network_code = Column(Text, nullable=True, index=True)
    network_region = Column(Text, index=True)
    network_name = Column(Text)

    active = Column(Boolean, default=True)

    dispatch_type = Column(
        Enum(DispatchType), nullable=False, default=DispatchType.GENERATOR
    )

    # @TODO remove when ref count is 0
    capacity_registered = Column(Numeric, nullable=True)

    registered = Column(DateTime, nullable=True)
    deregistered = Column(DateTime, nullable=True)

    unit_id = Column(Integer, nullable=True)
    unit_number = Column(Integer, nullable=True)
    unit_alias = Column(Text, nullable=True)
    unit_capacity = Column(Numeric, nullable=True)
    # unit_number_max = Column(Numeric, nullable=True)

    approved = Column(Boolean, default=False)
    approved_by = Column(Text)
    approved_at = Column(DateTime(timezone=True), nullable=True)

    @hybrid_property
    def capacity_aggregate(self) -> Optional[int]:
        """
            This is unit_no * unit_capacity and can differ from registered

        """
        num_units = 1
        cap_aggr = None

        if not self.active:
            return 0

        if self.unit_number and type(self.unit_number) is int:
            num_units = self.unit_number

        if self.unit_capacity and type(self.unit_capacity) is Decimal:
            cap_aggr = num_units * self.unit_capacity

        if type(cap_aggr) is Decimal:
            cap_aggr = round(cap_aggr, 2)

        return cap_aggr

    @hybrid_property
    def duid(self) -> str:
        return self.network_code or self.code

    @hybrid_property
    def status_label(self) -> Optional[str]:
        return self.status.label if self.status else None

    @hybrid_property
    def fueltech_label(self) -> Optional[str]:
        return self.fueltech.label if self.fueltech else None

    @hybrid_property
    def oid(self) -> str:
        return get_oid(self)

    @hybrid_property
    def ocode(self) -> str:
        return get_ocode(self)


class Revision(Base, BaseModel):

    __tablename__ = "revisions"

    id = Column(Integer, autoincrement=True, nullable=False, primary_key=True,)

    station_id = Column(
        Integer,
        ForeignKey("station.id", name="fk_revision_station_id"),
        nullable=True,
    )
    station = relationship(
        "Station", back_populates="revisions", lazy="joined"
    )

    facility_id = Column(
        Integer,
        ForeignKey("facility.id", name="fk_revision_facility_id"),
        nullable=True,
    )
    facility = relationship(
        "Facility", back_populates="revisions", lazy="joined"
    )

    location_id = Column(
        Integer,
        ForeignKey("location.id", name="fk_revision_location_id"),
        nullable=True,
    )
    location = relationship(
        "Location", back_populates="revisions", lazy="joined"
    )

    changes = Column(JSON, nullable=True)
    previous = Column(JSON, nullable=True)

    is_update = Column(Boolean, default=False)

    approved = Column(Boolean, default=False)
    approved_by = Column(Text)
    approved_at = Column(DateTime(timezone=True), nullable=True)
    approved_comment = Column(Text, nullable=True)

    discarded = Column(Boolean, default=False)
    discarded_by = Column(Text)
    discarded_at = Column(DateTime(timezone=True), nullable=True)

    @hybrid_property
    def parent_id(self) -> str:
        return self.station_id or self.facility_id or self.location_id

    @hybrid_property
    def parent_type(self) -> str:
        if self.station_id:
            return "station"

        if self.facility_id:
            return "facility"

        if self.location_id:
            return "location"

        return ""

    @hybrid_property
    def station_owner_id(self) -> int:
        if self.station_id:
            return self.station_id

        if self.facility_id:
            return self.facility.station.id

        if self.location:
            return self.location.station.id

    @hybrid_property
    def station_owner_name(self) -> str:
        if self.station_id:
            return self.station.name

        if self.facility_id:
            return self.facility.station.name

        if self.location:
            return self.location.station.name

    @hybrid_property
    def station_owner_code(self) -> str:
        if self.station_id:
            return self.station.code

        if self.facility_id:
            return self.facility.station.code

        if self.location:
            return self.location.station.code


class FacilityScada(Base, BaseModel):
    """
        Facility Scada
    """

    __tablename__ = "facility_scada"

    __table_args__ = (
        Index(
            "idx_facility_scada_trading_interval_year",
            text("date_trunc('year', trading_interval AT TIME ZONE 'UTC')"),
        ),
        Index(
            "idx_facility_scada_trading_interval_month",
            text("date_trunc('month', trading_interval AT TIME ZONE 'UTC')"),
        ),
        Index(
            "idx_facility_scada_trading_interval_day",
            text("date_trunc('day', trading_interval AT TIME ZONE 'UTC')"),
        ),
        Index(
            "idx_facility_scada_trading_interval_hour",
            text("date_trunc('hour', trading_interval AT TIME ZONE 'UTC')"),
        ),
        # new timezone based indicies
        # @NOTE: other indicies in migration files
    )

    def __str__(self) -> str:
        return "<{}: {} {} {}>".format(
            self.__class__,
            self.trading_interval,
            self.network_id,
            self.facility_code,
        )

    def __repr__(self) -> str:
        return "{}: {} {} {}".format(
            self.__class__,
            self.trading_interval,
            self.network_id,
            self.facility_code,
        )

    network_id = Column(
        Text,
        ForeignKey("network.code", name="fk_balancing_summary_network_code"),
        primary_key=True,
        nullable=False,
    )
    network = relationship("Network")

    trading_interval = Column(
        TIMESTAMP(timezone=True), index=True, primary_key=True, nullable=False
    )

    facility_code = Column(Text, nullable=False, primary_key=True, index=True)
    generated = Column(Numeric, nullable=True)
    eoi_quantity = Column(Numeric, nullable=True)


class BalancingSummary(Base, BaseModel):

    __tablename__ = "balancing_summary"

    __table_args__ = (
        Index(
            "idx_balancing_summary_trading_interval_year",
            text("date_trunc('year', trading_interval AT TIME ZONE 'UTC')"),
        ),
        Index(
            "idx_balancing_summary_trading_interval_month",
            text("date_trunc('month', trading_interval AT TIME ZONE 'UTC')"),
        ),
        Index(
            "idx_balancing_summary_trading_interval_day",
            text("date_trunc('day', trading_interval AT TIME ZONE 'UTC')"),
        ),
        Index(
            "idx_balancing_summary_trading_interval_hour",
            text("date_trunc('hour', trading_interval AT TIME ZONE 'UTC')"),
        ),
    )

    network_id = Column(
        Text,
        ForeignKey("network.code", name="fk_balancing_summary_network_code"),
        primary_key=True,
    )
    network = relationship("Network")

    trading_interval = Column(
        TIMESTAMP(timezone=True), index=True, primary_key=True
    )
    network_region = Column(Text, primary_key=True)
    forecast_load = Column(Numeric, nullable=True)
    generation_scheduled = Column(Numeric, nullable=True)
    generation_non_scheduled = Column(Numeric, nullable=True)
    generation_total = Column(Numeric, nullable=True)
    price = Column(Numeric, nullable=True)




import re
from typing import Optional

from opennem.core.normalizers import normalize_duid
from opennem.core.unit_parser import UnitSchema


def get_unit_code(
    unit: UnitSchema,
    duid: Optional[str] = None,
    station_name: Optional[str] = None,
) -> str:
    """
        This takes the duid from the network and the unit info and creates a unique
        opennem code

        This should be unique across all units for a network
    """

    unit_id = None
    # @TODO - check if we can strip, probably not a good idea

    if not duid:

        if not station_name:
            raise Exception(
                "Cannot generate a unit code without both duid and station name"
            )

        duid = get_basecode(station_name)

    duid_clean = normalize_duid(duid)

    if not duid_clean.endswith(str(unit.id)):
        unit_id = str(unit.id)

    if unit_id == 1:
        unit_id = None

    components = [duid_clean, unit.alias, unit_id]

    # empty out None's
    components = [i for i in components if i]

    unit_code = "_".join(components)

    return unit_code


OPENNEM_CODE_PREFIX = "0N"


__alphanum_strip = re.compile(r"[\W_\ ]+")


def clean_name_for_basecode(station_name: str) -> str:
    """
        Cleans up the name in prep to generate a basecode
    """

    return __alphanum_strip.sub(" ", station_name.strip())


def get_basecode(station_name: str) -> str:
    """
        Generate a code from the station name when there isn't an DUID

        We prefix these with 0N
    """

    if not type(station_name) is str or not station_name:
        raise Exception("Expected to generate a code with no station name")

    MIN_DUID_LENGTH = 4
    MAX_DUID_LENGTH = 6

    station_name = clean_name_for_basecode(station_name)

    comps = station_name.split(" ")

    num_words = len(comps)

    if num_words > MAX_DUID_LENGTH:
        comps = comps[:MAX_DUID_LENGTH]

    if num_words > MIN_DUID_LENGTH:
        num_words = MIN_DUID_LENGTH

    comps = [i[0 : MIN_DUID_LENGTH + 1 - num_words] for i in comps]

    comps.insert(0, OPENNEM_CODE_PREFIX)

    basecode = "".join(comps).upper()

    return basecode


def duid_is_ocode(duid: str) -> bool:
    return duid.startswith(OPENNEM_CODE_PREFIX)



import logging
import os
from datetime import datetime

from smart_open import open

from opennem.api.stats.router import (
    energy_network_fueltech_api,
    power_network_fueltech_api,
    price_network_region_api,
)
from opennem.api.weather.router import station_observations_api
from opennem.db import get_database_engine

YEAR_MIN = 2010

BASE_EXPORT = "s3://data.opennem.org.au"

BASE_EXPORT_LOCAL = os.path.realpath(
    os.path.join(os.path.dirname(__file__), "..", "..", "data")
)


UPLOAD_ARGS = {
    "ContentType": "application/json",
}

logger = logging.getLogger(__name__)


def wem_export_power():
    engine = get_database_engine()

    stat_set = power_network_fueltech_api(
        network_code="WEM",
        network_region="WEM",
        interval="30m",
        period="7d",
        engine=engine,
    )

    weather = station_observations_api(
        station_code="009021",
        interval="10m",
        period="7d",
        network_code="WEM",
        engine=engine,
    )

    price = price_network_region_api(
        engine=engine,
        network_code="WEM",
        region_code="WEM",
        interval="30m",
        period="7d",
    )

    # demand = wem_demand()

    stat_set.data = stat_set.data + price.data
    stat_set.data = stat_set.data + weather.data

    power_path = BASE_EXPORT + "/power/wem.json"

    with open(
        power_path,
        "w",
        transport_params=dict(multipart_upload_kwargs=UPLOAD_ARGS),
    ) as fh:
        fh.write(stat_set.json())


def wem_export_years():
    engine = get_database_engine()

    for year in range(datetime.now().year, YEAR_MIN - 1, -1):

        stat_set = None

        try:
            stat_set = energy_network_fueltech_api(
                network_code="WEM",
                network_region="WEM",
                interval="1d",
                year=year,
                period="1Y",
                engine=engine,
            )
        except Exception as e:
            logger.error(
                "Could not get energy network fueltech for {} {} : {}".format(
                    "WEM", year, e
                )
            )
            continue

        # market_value = wem_market_value_year(year)

        year_path = BASE_EXPORT + f"/wem/energy/daily/{year}.json"

        with open(
            year_path,
            "w",
            transport_params=dict(multipart_upload_kwargs=UPLOAD_ARGS),
        ) as fh:
            fh.write(stat_set.json())


def wem_export_all():

    engine = get_database_engine()

    stat_set = energy_network_fueltech_api(
        network_code="WEM",
        network_region="WEM",
        interval="1M",
        period="10Y",
        engine=engine,
    )

    # market_value = wem_market_value_all()

    all_path = BASE_EXPORT + "/wem/energy/monthly/all.json"

    with open(
        all_path,
        "w",
        transport_params=dict(multipart_upload_kwargs=UPLOAD_ARGS),
    ) as fh:
        fh.write(stat_set.json())


def wem_run_all():
    wem_export_power()
    wem_export_years()
    wem_export_all()


if __name__ == "__main__":
    wem_run_all()



import sentry_sdk
from pkg_resources import get_distribution

from opennem.settings import settings

# VERSION = get_distribution("opennem").version
# __version__ = VERSION


if settings.env != "development":
    sentry_sdk.init(
        settings.sentry_url, traces_sample_rate=1.0, environment=settings.env
    )




"""Balancing summary add network_region column

Revision ID: 911efdc0c77f
Revises: 8bf30c37cb85
Create Date: 2020-10-07 11:16:51.578193

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "911efdc0c77f"
down_revision = "8bf30c37cb85"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "balancing_summary",
        sa.Column("network_region", sa.Text(), nullable=True),
    )
    op.execute("COMMIT")
    op.execute(
        "update balancing_summary set network_region='WEM' where network_id='WEM'"
    )


def downgrade():
    op.drop_column("balancing_summary", "network_region")



"""

NEMWEB Data ingress into OpenNEM format


"""

import logging

from sqlalchemy.dialects.postgresql import insert

from opennem.core.normalizers import normalize_duid
from opennem.db import SessionLocal, get_database_engine
from opennem.db.models.opennem import BalancingSummary, FacilityScada
from opennem.schema.opennem import NetworkNEM
from opennem.utils.dates import parse_date
from opennem.utils.pipelines import check_spider_pipeline

logger = logging.getLogger(__name__)


def process_case_solutions(table):
    pass
    # session = SessionLocal()

    # if "records" not in table:
    #     raise Exception("Invalid table no records")

    # records = table["records"]

    # records_to = []


def process_meter_data_gen_duid(table, spider):
    session = SessionLocal()
    engine = get_database_engine()

    if "records" not in table:
        raise Exception("Invalid table no records")

    records = table["records"]

    records_to_store = []

    for record in records:
        trading_interval = parse_date(
            record["INTERVAL_DATETIME"], network=NetworkNEM, dayfirst=False
        )

        if not trading_interval:
            continue

        records_to_store.append(
            {
                "network_id": "NEM",
                "created_by": spider.name,
                # "updated_by": None,
                "trading_interval": trading_interval,
                "facility_code": normalize_duid(record["DUID"]),
                "eoi_quantity": record["MWH_READING"],
            }
        )

    stmt = insert(FacilityScada).values(records_to_store)
    stmt.bind = engine
    stmt = stmt.on_conflict_do_update(
        index_elements=["trading_interval", "network_id", "facility_code"],
        set_={"eoi_quantity": stmt.excluded.eoi_quantity},
    )

    try:
        session.execute(stmt)
        session.commit()
    except Exception as e:
        logger.error("Error inserting records")
        logger.error(e)
        return 0
    finally:
        session.close()

    return len(records_to_store)


def process_pre_ap_price(table, spider):
    session = SessionLocal()
    engine = get_database_engine()

    if "records" not in table:
        raise Exception("Invalid table no records")

    records = table["records"]

    records_to_store = []

    for record in records:
        trading_interval = parse_date(
            record["SETTLEMENTDATE"], network=NetworkNEM, dayfirst=False
        )

        if not trading_interval:
            continue

        records_to_store.append(
            {
                "network_id": "NEM",
                "created_by": spider.name,
                # "updated_by": None,
                "network_region": record["REGIONID"],
                "trading_interval": trading_interval,
                "price": record["PRE_AP_ENERGY_PRICE"],
            }
        )

    stmt = insert(BalancingSummary).values(records_to_store)
    stmt.bind = engine
    stmt = stmt.on_conflict_do_update(
        index_elements=["trading_interval", "network_id", "network_region"],
        set_={"price": stmt.excluded.price},
    )

    try:
        session.execute(stmt)
        session.commit()
    except Exception as e:
        logger.error("Error inserting records")
        logger.error(e)
        return 0
    finally:
        session.close()

    return len(records_to_store)


def process_unit_scada(table, spider):
    session = SessionLocal()
    engine = get_database_engine()

    if "records" not in table:
        raise Exception("Invalid table no records")

    records = table["records"]

    records_to_store = []
    records_primary_keys = []

    for record in records:
        trading_interval = parse_date(
            record["SETTLEMENTDATE"], network=NetworkNEM, dayfirst=False
        )
        facility_code = normalize_duid(record["DUID"])

        if not trading_interval or not facility_code:
            continue

        # Since this can insert 1M+ records at a time we need to
        # do a separate in-memory check of primary key constraints
        # better way of doing this .. @TODO
        _unique_set = (trading_interval, facility_code, "NEM")

        if _unique_set not in records_primary_keys:

            records_to_store.append(
                {
                    "network_id": "NEM",
                    "created_by": spider.name,
                    # "updated_by": None,
                    "trading_interval": trading_interval,
                    "facility_code": facility_code,
                    "generated": float(record["SCADAVALUE"]),
                }
            )

            records_primary_keys.append(_unique_set)

    # free
    records_primary_keys = []

    logger.debug("Saving %d records", len(records_to_store))

    stmt = insert(FacilityScada).values(records_to_store)
    stmt.bind = engine
    stmt = stmt.on_conflict_do_update(
        index_elements=["trading_interval", "network_id", "facility_code"],
        set_={"generated": stmt.excluded.generated},
    )

    try:
        session.execute(stmt)
        session.commit()
    except Exception as e:
        logger.error("Error: {}".format(e))
        return 0
    finally:
        session.close()

    return len(records_to_store)


def process_unit_solution(table, spider):
    session = SessionLocal()
    engine = get_database_engine()

    if "records" not in table:
        raise Exception("Invalid table no records")

    records = table["records"]

    records_to_store = []
    records_primary_keys = []

    for record in records:
        trading_interval = parse_date(
            record["SETTLEMENTDATE"], network=NetworkNEM, dayfirst=False
        )
        facility_code = normalize_duid(record["DUID"])

        if not trading_interval or not facility_code:
            continue

        # Since this can insert 1M+ records at a time we need to
        # do a separate in-memory check of primary key constraints
        # better way of doing this .. @TODO
        _unique_set = (trading_interval, facility_code, "NEM")

        if _unique_set not in records_primary_keys:

            records_to_store.append(
                {
                    "network_id": "NEM",
                    "created_by": spider.name,
                    # "updated_by": None,
                    "trading_interval": trading_interval,
                    "facility_code": facility_code,
                    "eoi_quantity": float(record["INITIALMW"]),
                }
            )
            records_primary_keys.append(_unique_set)

    # free
    records_primary_keys = []

    logger.debug("Saving %d records", len(records_to_store))

    stmt = insert(FacilityScada).values(records_to_store)
    stmt.bind = engine
    stmt = stmt.on_conflict_do_update(
        index_elements=["trading_interval", "network_id", "facility_code"],
        set_={"eoi_quantity": stmt.excluded.eoi_quantity},
    )

    try:
        session.execute(stmt)
        session.commit()
    except Exception as e:
        logger.error("Error: {}".format(e))
        return 0
    finally:
        session.close()

    return len(records_to_store)


TABLE_PROCESSOR_MAP = {
    "METER_DATA_GEN_DUID": "process_meter_data_gen_duid",
    "DISPATCH_CASE_SOLUTION": "process_case_solutions",
    "DISPATCH_UNIT_SCADA": "process_unit_scada",
    "DISPATCH_UNIT_SOLUTION": "process_unit_solution",
    "DISPATCH_PRE_AP_PRICE": "process_pre_ap_price",
}


class NemwebUnitScadaOpenNEMStorePipeline(object):
    @check_spider_pipeline
    def process_item(self, item, spider=None):

        if "tables" not in item:
            logger.error("Invalid item - no tables located")
            return 0

        if not isinstance(item["tables"], list):
            logger.error("Invalid item - no tables located")
            return 0

        tables = item["tables"]

        ret = 0

        for table in tables:
            if "name" not in table:
                logger.info("Invalid table found")
                continue

            table_name = table["name"]

            if table_name not in TABLE_PROCESSOR_MAP:
                logger.error("No processor for table %s", table_name)
                continue

            process_meth = TABLE_PROCESSOR_MAP[table_name]

            if process_meth not in globals():
                logger.info("Invalid processing function %s", process_meth)
                continue

            ret += globals()[process_meth](table, spider=spider)

        return ret



import json

import scrapy

from opennem.pipelines.bom import StoreBomObservation


class BomJSONObservationSpider(scrapy.Spider):
    allowed_domains = ["bom.gov.au"]

    start_urls = ["http://reg.bom.gov.au/fwo/IDN60901/IDN60901.94767.json"]

    pipelines = set([StoreBomObservation])

    station_id: str = None

    def parse(self, response):
        json_response = json.loads(response.text)

        if not type(json_response) is dict:
            raise Exception("invalid response from bom")

        if "observations" not in json_response:
            raise Exception("invalid response from bom")

        if "data" not in json_response["observations"]:
            raise Exception("invalid response from bom")

        code = None

        if "code" in response.meta:
            code = response.meta["code"]

        elif hasattr(self, "station_id") and self.station_id:
            code = self.station_id

        if not code:
            raise Exception("No station id for this scrape can't join")

        records = []

        for i in json_response["observations"]["data"]:
            i["code"] = code
            records.append(i)

        yield {"records": records}



import logging
from typing import Generator

from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

from opennem.exporter.encoders import opennem_deserialize, opennem_serialize
from opennem.settings import settings

DeclarativeBase = declarative_base()

logger = logging.getLogger(__name__)


def db_connect(db_name=None, debug=False):
    """
    Performs database connection using database settings from settings.py.

    Returns sqlalchemy engine instance
    """
    db_conn_str = settings.db_url

    connect_args = {}

    if db_conn_str.startswith("sqlite"):
        connect_args = {"check_same_thread": False}

    try:
        return create_engine(
            db_conn_str,
            json_serializer=opennem_serialize,
            json_deserializer=opennem_deserialize,
            echo=debug,
            pool_size=10,
            pool_timeout=60,
            connect_args=connect_args,
        )
    except Exception as exc:
        logger.error("Could not connect to database: %s", exc)


engine = db_connect()
SessionLocal = sessionmaker(bind=engine, autocommit=False, autoflush=False,)
SessionAutocommit = sessionmaker(bind=engine, autocommit=True, autoflush=True)


def get_database_session() -> Generator:
    """
        Gets a database session

    """
    try:
        s = SessionLocal()
        yield s
    except Exception as e:
        raise e
    finally:
        s.close()


def get_database_engine():
    """
        Gets a database engine connection

    """
    return engine



from sqlalchemy import (
    DECIMAL,
    Column,
    Date,
    ForeignKey,
    Index,
    String,
    Table,
    text,
)
from sqlalchemy.dialects.mysql import (
    INTEGER,
    MEDIUMINT,
    SMALLINT,
    TINYINT,
    VARCHAR,
)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

Base = declarative_base()
metadata = Base.metadata


class CATEGORY(Base):
    __tablename__ = "CATEGORY"

    ID = Column(TINYINT(4), primary_key=True)
    CATEGORY = Column(String(10, "utf8_bin"), nullable=False, unique=True)


class CLASSIFICATION(Base):
    __tablename__ = "CLASSIFICATION"

    ID = Column(TINYINT(4), primary_key=True)
    CLASSIFICATION = Column(String(15, "utf8_bin"), unique=True)


t_CURRENT_REGISTER = Table(
    "CURRENT_REGISTER",
    metadata,
    Column("ID", SMALLINT(6), server_default=text("'0'")),
    Column("DUID", String(10)),
    Column("STATION_NAME", String(100)),
    Column("PARTICIPANT", String(80)),
    Column("CLASSIFICATION", String(15)),
    Column("CATEGORY", String(10)),
    Column("TECHNOLOGY_TYPE", String(60)),
    Column("FUEL_SOURCE", String(45)),
    Column("REGIONID", String(6)),
    Column("DISPATCH_TYPE", String(45)),
    Column("UNIT_NO", String(50)),
    Column("UNIT_SIZE", DECIMAL(7, 3)),
    Column("REG_CAP", DECIMAL(7, 3)),
    Column("MAX_CAP", DECIMAL(7, 3)),
    Column("MAX_ROC", DECIMAL(4, 0)),
    Column("DATE", Date),
)


t_CURRENT_REGISTER_DATE = Table(
    "CURRENT_REGISTER_DATE", metadata, Column("DATE", Date)
)


class DISPATCHTYPE(Base):
    __tablename__ = "DISPATCH_TYPE"

    ID = Column(TINYINT(4), primary_key=True)
    DISPATCH_TYPE = Column(String(45, "utf8_bin"), nullable=False, unique=True)


class DUID(Base):
    __tablename__ = "DUID"

    ID = Column(SMALLINT(6), primary_key=True)
    DUID = Column(VARCHAR(10), unique=True)


t_DUID_EFFICIENCY = Table(
    "DUID_EFFICIENCY",
    metadata,
    Column("DUID", SMALLINT(6)),
    Column("THERMAL_EFFICIENCY", DECIMAL(7, 4)),
)


class FUELCATEGORY(Base):
    __tablename__ = "FUEL_CATEGORY"

    ID = Column(TINYINT(4), primary_key=True, unique=True)
    FUEL_CATEGORY = Column(String(45, "utf8_bin"))


class FUELTECH(Base):
    __tablename__ = "FUEL_TECHS"

    ID = Column(TINYINT(4), primary_key=True)
    FUEL_TECH = Column(String(45, "utf8_bin"), unique=True)
    openNEM_keys = Column(String(45, "utf8_bin"), unique=True)
    openNEM_colors = Column(String(7, "utf8_bin"))


t_FULL_REGISTER = Table(
    "FULL_REGISTER",
    metadata,
    Column("ID", SMALLINT(6), server_default=text("'0'")),
    Column("DUID", String(10)),
    Column("STATION_NAME", String(100)),
    Column("PARTICIPANT", String(80)),
    Column("CLASSIFICATION", String(15)),
    Column("CATEGORY", String(10)),
    Column("DISPATCH_TYPE", String(45)),
    Column("REGIONID", String(6)),
    Column("FUEL_CATEGORY", String(45)),
    Column("TECHNOLOGY_CATEGORY", String(45)),
    Column("FUEL_TECH", String(45)),
    Column("UNIT_NO", String(50)),
    Column("UNIT_SIZE", DECIMAL(7, 3)),
    Column("REG_CAP", DECIMAL(7, 3)),
    Column("MAX_CAP", DECIMAL(7, 3)),
    Column("MAX_ROC", DECIMAL(4, 0)),
    Column("DATE", Date),
)


t_GENERATOR_TECHNICAL_DATA = Table(
    "GENERATOR_TECHNICAL_DATA",
    metadata,
    Column("STATION_NAME_ID", SMALLINT(6)),
    Column("STATION_NAME", String(100)),
    Column("REGIONID", String(6)),
    Column("FUEL_TECH", String(45)),
    Column("CAPACITY", DECIMAL(29, 3)),
    Column("THERMAL_EFFICIENCY", DECIMAL(7, 4)),
    Column("SCOPE_1", DECIMAL(5, 4)),
)


t_LAST_REGISTERED_DATE = Table(
    "LAST_REGISTERED_DATE",
    metadata,
    Column("DUID", SMALLINT(6)),
    Column("LAST_DATE", Date),
)


t_NEM_EMISSION_INTENSITY = Table(
    "NEM_EMISSION_INTENSITY",
    metadata,
    Column("ID", SMALLINT(6), server_default=text("'0'")),
    Column("STATION_NAME", String(100)),
    Column("SCOPE_1", DECIMAL(5, 4)),
    Column("SCOPE_3", DECIMAL(5, 4)),
    Column("TOTAL", DECIMAL(6, 4)),
)


class PARTICIPANT(Base):
    __tablename__ = "PARTICIPANTS"

    ID = Column(SMALLINT(6), primary_key=True)
    PARTICIPANT = Column(String(80, "utf8_bin"), unique=True)


class REGIONID(Base):
    __tablename__ = "REGIONID"

    id = Column(TINYINT(4), primary_key=True)
    REGIONID = Column(VARCHAR(6), unique=True)


t_REGION_SUMMARY = Table(
    "REGION_SUMMARY",
    metadata,
    Column("REGIONID", String(6)),
    Column("FUEL_TECH", String(45)),
    Column("MW", DECIMAL(29, 3)),
)


t_REGISTERED_GENERATORS = Table(
    "REGISTERED_GENERATORS",
    metadata,
    Column("STATION_NAME", String(100)),
    Column("MW", DECIMAL(29, 3)),
)


t_STANDARDISED_REGISTER = Table(
    "STANDARDISED_REGISTER",
    metadata,
    Column("id", SMALLINT(6), server_default=text("'0'")),
    Column("DUID", String(10)),
    Column("STATION_NAME", String(100)),
    Column("PARTICIPANT", String(80)),
    Column("CLASSIFICATION", String(15)),
    Column("CATEGORY", String(10)),
    Column("DISPATCH_TYPE", String(45)),
    Column("REGIONID", String(6)),
    Column("FUEL_CATEGORY", String(45)),
    Column("TECHNOLOGY_CATEGORY", String(45)),
    Column("FUEL_TECH", String(45)),
    Column("UNIT_NO", String(50)),
    Column("UNIT_SIZE", DECIMAL(7, 3)),
    Column("REG_CAP", DECIMAL(7, 3)),
    Column("MAX_CAP", DECIMAL(7, 3)),
    Column("MAX_ROC", DECIMAL(4, 0)),
    Column("DATE", Date),
)


class STATIONNAME(Base):
    __tablename__ = "STATION_NAMES"

    ID = Column(SMALLINT(6), primary_key=True)
    STATION_NAME = Column(String(100, "utf8_bin"), unique=True)


t_STATION_SUMMARY = Table(
    "STATION_SUMMARY",
    metadata,
    Column("ID", SMALLINT(6), server_default=text("'0'")),
    Column("DUID", String(10)),
    Column("STATION_NAME_ID", SMALLINT(6), server_default=text("'0'")),
    Column("STATION_NAME", String(100)),
    Column("PARTICIPANT", String(80)),
    Column("REGIONID", String(6)),
    Column("openNEM_keys", String(45)),
    Column("openNEM_colors", String(7)),
    Column("REG_CAP", DECIMAL(7, 3)),
    Column("MAX_CAP", DECIMAL(7, 3)),
    Column("MAX_ROC", DECIMAL(4, 0)),
)


class TECHNOLOGYCATEGORY(Base):
    __tablename__ = "TECHNOLOGY_CATEGORY"

    ID = Column(TINYINT(4), primary_key=True)
    TECHNOLOGY_CATEGORY = Column(String(45, "utf8_bin"), unique=True)


t__MAPPING_FUEL_SOURCE = Table(
    "_MAPPING_FUEL_SOURCE",
    metadata,
    Column("ID", TINYINT(4), server_default=text("'0'")),
    Column("FUEL_SOURCE", String(45)),
    Column("FUEL_CATEGORY", String(45)),
)


t__MAPPING_TECHNOLOGY_TYPE = Table(
    "_MAPPING_TECHNOLOGY_TYPE",
    metadata,
    Column("ID", TINYINT(4), server_default=text("'0'")),
    Column("TECHNOLOGY_TYPE", String(60)),
    Column("TECHNOLOGY_CATEGORY", String(45)),
)


t__MAPPING_TECH_FUEL = Table(
    "_MAPPING_TECH_FUEL",
    metadata,
    Column("FUEL_TECH", String(45)),
    Column("FUEL_CATEGORY", String(45)),
    Column("TECHNOLOGY_CATEGORY", String(45)),
)


class FUELSOURCE(Base):
    __tablename__ = "FUEL_SOURCES"

    ID = Column(TINYINT(4), primary_key=True)
    FUEL_SOURCE = Column(String(45, "utf8_bin"), unique=True)
    FUEL_CATEGORY_ID = Column(ForeignKey("FUEL_CATEGORY.ID"), index=True)

    FUEL_CATEGORY = relationship("FUELCATEGORY")


class FUELTECHMAPPING(Base):
    __tablename__ = "FUEL_TECH_MAPPING"

    ID = Column(TINYINT(4), primary_key=True)
    FUEL_CATEGORY_ID = Column(ForeignKey("FUEL_CATEGORY.ID"), index=True)
    TECHNOLOGY_CATEGORY_ID = Column(
        ForeignKey("TECHNOLOGY_CATEGORY.ID"), index=True
    )
    FUEL_TECH_ID = Column(ForeignKey("FUEL_TECHS.ID"), index=True)

    FUEL_CATEGORY = relationship("FUELCATEGORY")
    FUEL_TECH = relationship("FUELTECH")
    TECHNOLOGY_CATEGORY = relationship("TECHNOLOGYCATEGORY")


class INTERCONNECTORID(Base):
    __tablename__ = "INTERCONNECTORID"

    id = Column(TINYINT(4), primary_key=True)
    INTERCONNECTORID = Column(VARCHAR(9), nullable=False, unique=True)
    FROM_REGIONID = Column(ForeignKey("REGIONID.id"), index=True)
    TO_REGIONID = Column(ForeignKey("REGIONID.id"), index=True)

    REGIONID = relationship(
        "REGIONID", primaryjoin="INTERCONNECTORID.FROM_REGIONID == REGIONID.id"
    )
    REGIONID1 = relationship(
        "REGIONID", primaryjoin="INTERCONNECTORID.TO_REGIONID == REGIONID.id"
    )


class NTNDPEMISSION(Base):
    __tablename__ = "NTNDP_EMISSIONS"

    ID = Column(INTEGER(11), primary_key=True)
    NTNDP_STATION_NAME = Column(String(45, "utf8_bin"), unique=True)
    STATION_NAME_ID = Column(ForeignKey("STATION_NAMES.ID"), unique=True)
    SCOPE_1 = Column(DECIMAL(5, 4))
    SCOPE_3 = Column(DECIMAL(5, 4))

    STATION_NAME = relationship("STATIONNAME")


class NTNDPTECHNICALDATUM(Base):
    __tablename__ = "NTNDP_TECHNICAL_DATA"

    ID = Column(INTEGER(11), primary_key=True)
    UNIT = Column(String(45, "utf8_bin"))
    DUID = Column(ForeignKey("DUID.ID"), index=True)
    FIRST_RUN = Column(SMALLINT(6))
    AUX_LOAD = Column(TINYINT(4))
    HOT_DAY_DERATE = Column(DECIMAL(4, 2))
    ROCUP_START = Column(SMALLINT(6))
    ROCUP_NORM = Column(SMALLINT(6))
    ROCDOWN_START = Column(SMALLINT(6))
    ROCDOWN_NORM = Column(MEDIUMINT(9))
    THERMAL_EFFICIENCY = Column(TINYINT(4))

    DUID1 = relationship("DUID")


t_RET_BASELINE = Table(
    "RET_BASELINE",
    metadata,
    Column("ID", ForeignKey("DUID.ID"), nullable=False, index=True),
    Column("BASELINE", INTEGER(11), nullable=False),
)


class TECHNOLOGYTYPE(Base):
    __tablename__ = "TECHNOLOGY_TYPES"

    ID = Column(TINYINT(4), primary_key=True)
    TECHNOLOGY_TYPE = Column(String(60, "utf8_bin"), unique=True)
    TECHNOLOGY_CATEGORY_ID = Column(
        ForeignKey("TECHNOLOGY_CATEGORY.ID"), index=True
    )

    TECHNOLOGY_CATEGORY = relationship("TECHNOLOGYCATEGORY")


class REGISTER(Base):
    __tablename__ = "REGISTER"
    __table_args__ = (
        Index("uniq", "DATE", "DUID", "UNIT_SIZE", "UNIT_NO", unique=True),
    )

    ID = Column(INTEGER(11), primary_key=True, unique=True)
    DUID = Column(ForeignKey("DUID.ID"), nullable=False, index=True)
    PARTICIPANT_ID = Column(
        ForeignKey("PARTICIPANTS.ID"), nullable=False, index=True
    )
    STATION_NAME_ID = Column(
        ForeignKey("STATION_NAMES.ID"), nullable=False, index=True
    )
    REGIONID = Column(ForeignKey("REGIONID.id"), nullable=False, index=True)
    DISPATCH_TYPE_ID = Column(
        ForeignKey("DISPATCH_TYPE.ID"), nullable=False, index=True
    )
    CATEGORY_ID = Column(ForeignKey("CATEGORY.ID"), nullable=False, index=True)
    CLASSIFICATION_ID = Column(
        ForeignKey("CLASSIFICATION.ID"), nullable=False, index=True
    )
    FUEL_SOURCE_ID = Column(
        ForeignKey("FUEL_SOURCES.ID"), nullable=False, index=True
    )
    TECHNOLOGY_TYPE_ID = Column(
        ForeignKey("TECHNOLOGY_TYPES.ID"), nullable=False, index=True
    )
    UNIT_NO = Column(VARCHAR(50))
    UNIT_SIZE = Column(DECIMAL(7, 3))
    MAX_CAP = Column(DECIMAL(7, 3))
    REG_CAP = Column(DECIMAL(7, 3))
    MAX_ROC = Column(DECIMAL(4, 0))
    DATE = Column(Date, nullable=False, index=True)

    CATEGORY = relationship("CATEGORY")
    CLASSIFICATION = relationship("CLASSIFICATION")
    DISPATCH_TYPE = relationship("DISPATCHTYPE")
    DUID1 = relationship("DUID")
    FUEL_SOURCE = relationship("FUELSOURCE")
    PARTICIPANT = relationship("PARTICIPANT")
    REGIONID1 = relationship("REGIONID")
    STATION_NAME = relationship("STATIONNAME")
    TECHNOLOGY_TYPE = relationship("TECHNOLOGYTYPE")
